//################################################################# OSIALOD.C
// Purpose...: OSI File converter
// LastEdit..: 31.10.2022
// History...: Originally written by Mark Spankus
//             221031: Added -o switch, conversion from checksum load format
//                     to OSI QuickLoad format              
//###########################################################################

//================================================================== includes
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>

typedef unsigned char BYTE;
typedef unsigned char BOOL;
typedef unsigned int  WORD;

#define m_BUFSIZE                            256 // read buffer size
#define m_LOADER_TYPE_CS                       1 // Checksum loader 
#define m_LOADER_TYPE_QL                       2 // QuickLoad loader

int  m_iLineEnd = 0;                             // line end CRLF / CR
long m_lLoadAddr = 0;                            // 
long m_lGoAddr = -1;                             // where to go after load
int  m_iNullPadding = 5;                         // null padding

//==================================================================== Help()
// Purpose...: Display usage info
//===========================================================================
void Help(void) {

  char abBuf[] = {
    "\r\nOSIALOD V2.00 Usage:\r\n\r\n"
    "OSIALOD [-l<addr>][-g<addr>][-t][-f][-p<nn>]\r\n"
    "        [-a|-v|-c|-i[addr]|-c[addr]|-q[addr]] [InFile] [OutFile]\r\n\r\n"

    "-a        Convert binary to OSI 65A Serial Monitor Load Format\r\n"
    "-v        Convert binary to OSI 65V ROM Monitor Load Format\r\n"
    "-c        Convert binary to MOS/OSI Checksum Load Format\r\n"
    "-i[addr]  Convert binary to Checksum load format\r\n"
    "-q[addr]  Convert binary to OSI QSave Binary Load Format\r\n"
    "-o[addr]  Convert 65V ROM Monitor Load Format to QSave Format\r\n"
    "-l<addr>  Load/target address\r\n"
    "-g<addr>  Go/execute address, defaults to load address\r\n"
    "-p<nn>    Number of nul padding bytes at end of line, default 5\r\n"
    "-t        Use CR/LF at end of line, normally just CR\r\n"
    "-f        Force overwrite of output file\r\n\r\n"

    "InFile defaults to stdin, OutFile to stdout\r\n\r\n"

    "-i,-q,-o operations write a checksum loader prior to data\r\n"
    "[addr] specifies loader address, defaults to load address -0x100\r\n\r\n"

    "All values must be entered in hex notation\r\n\r\n"
  };

  fputs(abBuf, stdout);
}
//================================================================ eoF Help()

//*********************************************************** utility section
//========================================================== CnvHexStrToBin()
// Purpose...: Convert HEX-ASCII string to binary value
// Input.....: i_pcStr.: Address of HEX string
//             i_iWidth: Number of digits to convert
// Result....: Converted binary value
//===========================================================================
long CnvHexStrToBin(const char* i_pcStr, int i_iWidth) {
    
  char cHexCh;
  BOOL fInvalid;
  long lFnResult = 0;
	short sCount = 0;

	if ((i_pcStr != NULL) && (i_iWidth > 0) && (i_iWidth <= 8)) {

    while ((*i_pcStr) && (sCount < i_iWidth)) {
      cHexCh = *i_pcStr;

      fInvalid  = cHexCh == 0x20;
      fInvalid |= cHexCh == 0x0A;
      fInvalid |= cHexCh == 0x0D;
      fInvalid |= cHexCh == 0x09;

      if (! fInvalid) {
        lFnResult <<= 4;

        if ((cHexCh >= 'a') && (cHexCh <= 'f')) {
          cHexCh -= 0x20;
    		}

        if ((cHexCh >= 'A') && (cHexCh <= 'F')) {
          lFnResult |= (cHexCh - '7');
        }
    	  else if ((cHexCh >= '0') && (cHexCh <= '9')) {
          lFnResult |= (cHexCh - '0');
        }
    	  else {
          sCount = i_iWidth;
        }
      }

      else {
        sCount --;
      }

      i_pcStr ++;
      sCount ++;
    }
  }

  return lFnResult;
}
//====================================================== eoF CnvHexStrToBin()

//================================================================ WriteHex()
// Purpose...: Write byte as two-digit HEX to outfile
// Input.....: i_bData: Data byte to convert and write
//             i_pfOut: File pointer to output file
//===========================================================================
int WriteHex(BYTE i_bData, FILE* i_pfOut) {

  BYTE bHexCh;
  int iFnResult = 0;

  bHexCh = (i_bData >> 4);

  if (bHexCh > 9) {
    bHexCh = bHexCh + 7;
  }

  iFnResult = fputc(bHexCh +'0', i_pfOut);
  bHexCh = i_bData & '\x0F';

  if (bHexCh > 9) {
    bHexCh += 7;
  }

  if (iFnResult != EOF) {
    iFnResult = fputc(bHexCh +'0', i_pfOut);
  }

	return iFnResult;
}
//============================================================ eoF WriteHex()

//================================================================ WriteEol()
// Purpose...: Write end of line for OSI or PC
// Input.....: i_pfOut: File pointer to output file
// Result....: futc result
//===========================================================================
int WriteEol(FILE* i_pfOut) {

  int iFnResult = 0;

	switch(m_iLineEnd) {

		case -1:
			iFnResult = 0;
    break;
		
		case 0:
			iFnResult = fputc('\x0D', i_pfOut);
    break;

		default:
      case 1:
			  iFnResult = fputs("\x0D\x0A", i_pfOut);
			break;
  }

  return (iFnResult);
}
//============================================================ eoF WriteEol()
//******************************************************* end utility section

//********************************************************* QuickLoad section
//================================================================ WriteQGo()
// Purpose...: Write go for QSave
// Input.....: i_lAddr: Address to write
//             i_pfOut: File pointer to output file
//===========================================================================
int WriteQGo(long i_lAddr, FILE* i_pfOut) {
    
  int iFnResult = 0;

  if (i_lAddr >= 0) {

    //------------------------------------------------------ set address mode
    iFnResult = fputc( '$', i_pfOut);

    //--------------------------------------------------------- write address
    if (iFnResult != EOF) {
      iFnResult = fputc((BYTE)((i_lAddr >> 8) & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = fputc((BYTE)(i_lAddr & 0xFF), i_pfOut);
    }
  }

  return (iFnResult);
}
//============================================================ eoF WriteQGo()

//============================================================== WriteQSave()
// Purpose...: 8bit version of MOS/OSI checksum format (qsave, *.qsv)
//             ;<len><adrh><adrl><data bytes><chksum hi><chksm lo>
//             $<adrhi><adrlo> [EXECUTE]
// Input.....: i_pfIn.: File pointer input file
//             i_pfOut: File pointer output file
// Result....: fputc() result
//===========================================================================
int WriteQSave(FILE* i_pfIn, FILE* i_pfOut) {

  #define REC_SIZE 0x38

  WORD wSum;
  int iCnt;
  char acBuf[m_BUFSIZE];
  unsigned long ulAddr = m_lLoadAddr;
  int iNumRead = 0;
  int iFnResult = 0;

  while (iFnResult != EOF) {
    iNumRead = fread(acBuf, sizeof(char), REC_SIZE, i_pfIn);

    if (! iNumRead) {
      break;
    }

    //-------------------------------------------------- addr<data><cksm><CR>
    iFnResult = fputc(';', i_pfOut);
    wSum = iNumRead;

    //--------------------------------------- write number of bytes in record
    if (iFnResult != EOF) {
      iFnResult = fputc((BYTE)iNumRead, i_pfOut);
    }

    //---------------------------------------------------- write current addr
    if (iFnResult != EOF) {
      wSum += (BYTE)((ulAddr >> 8) & 0xFF);
      iFnResult = fputc((BYTE)((ulAddr >> 8) & 0xFF), i_pfOut);
      wSum += (BYTE)(ulAddr & 0xFF);

      if (iFnResult != EOF) {
        iFnResult = fputc((BYTE)(ulAddr & 0xFF), i_pfOut);
      }
    }

    ulAddr += iNumRead;

    //------------------------------------------------------ write data bytes
    for (iCnt = 0; ((iCnt < iNumRead) && (iFnResult != EOF)); iCnt ++) {
      iFnResult = fputc((BYTE)acBuf[iCnt], i_pfOut);
      wSum += (BYTE)acBuf[iCnt];
    }

    //-------------------------------------------------------- write checksum
    if (iFnResult != EOF) {
      iFnResult = fputc((BYTE)((wSum >> 8) & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = fputc((BYTE)(wSum & 0xFF), i_pfOut);
    }

    //-------------------------------------------------------- <CR> ends data
    if (iFnResult != EOF) {
      fputc('\r', i_pfOut);
    }

    //-------------------------------------------------------------- padding?
    for (iCnt = 0; ((iCnt < m_iNullPadding)&&(iFnResult != EOF)); iCnt ++) {
      iFnResult = fputc( '\x0', i_pfOut);
    }

    //------------------------------------------------- <LF> starts next line
    if (iFnResult != EOF) {
      fputc('\n', i_pfOut);
    }
  }

  fclose(i_pfIn);

  if ((m_lGoAddr >= 0) && (iFnResult != EOF)) {
    iFnResult = WriteQGo(m_lGoAddr, i_pfOut);
  }

  fclose(i_pfOut);
  return iFnResult;
}
//========================================================== eoF WriteQSave()

//============================================================== WriteQlRec()
// Purpose...: Write one record to QuickSave file
// Input.....: i_pfOut: File pointer to (opened) output file
//             i_wAddr: Address to apply
//             i_pbBuf: Address of buffer which contains data
//             i_iCnt.: Number of data bytes in record
//===========================================================================
int WriteQlRec(FILE* i_pfOut, WORD i_wAddr, BYTE* i_pbBuf, int i_iCnt) {

  int iCnt;
  BYTE bData;
  int iFnResult = 0;
  unsigned short wCheckSum = 0;

  //-------------------------------------------- write record start character
  iFnResult = fputc((BYTE)';', i_pfOut);

  //----------------------------------------------------- write record length
  if (iFnResult != EOF) {
    bData = (BYTE)i_iCnt;
    iFnResult = fputc(bData, i_pfOut);
    wCheckSum += bData;
  }

  //----------------------------------------------------------- write address
  if (iFnResult != EOF) {
    bData = (i_wAddr >> 8) & 0x00FF;
    fputc(bData, i_pfOut);
    wCheckSum += bData;
    bData = (i_wAddr & 0x00FF);
    iFnResult = fputc(bData, i_pfOut);
    wCheckSum += bData;
  }

  //-------------------------------------------------------- write data bytes
  if (iFnResult != EOF) {

    for (iCnt = 0; iCnt < i_iCnt; iCnt ++) {
      iFnResult = fputc(i_pbBuf[iCnt], i_pfOut);
      wCheckSum += i_pbBuf[iCnt];
      iCnt = (iFnResult == EOF) ? i_iCnt : iCnt;
    }
  }

  //---------------------------------------------------------- write checksum
  if (iFnResult != EOF) {
    bData = (wCheckSum >> 8) & 0x00FF;
    fputc(bData, i_pfOut);
    bData = (wCheckSum & 0x00FF);
    iFnResult = fputc (bData, i_pfOut);
  }

  return (iFnResult);
}
//========================================================== eoF WriteQlRec()

//========================================================== LodToQuickLoad()
// Purpose...: Read Checksum LOD file, convert non-loader part to binary,
//             then call WriteQSave to save it in QSAVE format
//===========================================================================
int LodToQuickLoad(FILE* i_pfIn, FILE* i_pfOut) {

  #define QLRECSIZE 0x38

  BYTE abBuf[4] = {0};
  BYTE abRecBuf[0x40] = {0};
  BYTE* pbRecBuf = abRecBuf;
  WORD wLodAddr = 0;
  WORD wQlAddr = 0;
  int iCnt = 0;
  int iQlRecCnt = 0;
  int iLodRecSize = 0;
  int iNumRead = 1;
  int iFnResult = 1;

  //-------------------------------------------------------------------------
  // Write binary mark immediately after 'G'<address> for OSI.COM on Kaypro
  // to recognize the file as binary. Should not harm loading from tape
  //-------------------------------------------------------------------------
  fputc('\x7F', i_pfOut);

  while (iNumRead != EOF) {

    //---------------------- search for ';', the start of the lod data record
    while ((iNumRead != 0) && (abBuf[0] != ';') && (abBuf[0] != '$')) {
      iNumRead = fread(&abBuf, sizeof(char), 1, i_pfIn);
    }

    //--------------------------------------------------------- go-mark found
    if ((iNumRead == 1) && (abBuf[0] == '$')) {
      WriteQlRec(i_pfOut, wQlAddr, pbRecBuf, iQlRecCnt);
      iNumRead = fread(&abBuf, sizeof(char), 4, i_pfIn);

      if (iNumRead == 4) {
        long lGoAddr = CnvHexStrToBin(abBuf, 4);
        WriteQGo(lGoAddr, i_pfOut);
        iNumRead = EOF;
      }
    }

    //-------------------------------------------------- read lod data record
    else if ((iNumRead == 1) && (abBuf[0] == ';')) {

      //----------------------------------------------------- get record size
      iNumRead = fread(&abBuf, sizeof(char), 2, i_pfIn);

      if (iNumRead == 2) {
        iLodRecSize = CnvHexStrToBin(abBuf, 2);
      }

      //--------------------------------------------------------- get address
      if ((iLodRecSize != 0) && (iLodRecSize <= 24)) {
        iNumRead = fread(&abBuf, sizeof(char), 4, i_pfIn);

        if (iNumRead == 4) {
          wLodAddr = CnvHexStrToBin(abBuf, 4);

          if (wQlAddr == 0) {
            wQlAddr = wLodAddr;
          }
        }
      }

      //-------------------------------------------------- process data bytes
      for (iCnt = 0; iCnt < iLodRecSize; iCnt ++) {
        iNumRead = fread(&abBuf, sizeof(char), 2, i_pfIn);

        if (iNumRead == 2) {
          pbRecBuf[iQlRecCnt] = CnvHexStrToBin(abBuf, 2);
          iQlRecCnt ++;

          //------------ check if ql recsize reached, if so, write to outfile
          if (iQlRecCnt == QLRECSIZE) {
            WriteQlRec(i_pfOut, wQlAddr, pbRecBuf, iQlRecCnt);
            wQlAddr += iQlRecCnt;
            iQlRecCnt = 0;
          }
        }
      }

      //------------------------------------ read checksum (not checked here)
      if (iNumRead == 2) {
        iNumRead = fread(&abBuf, sizeof(char), 2, i_pfIn);
      }
    }
  }

  return (iFnResult);
}
//====================================================== eoF LodToQuickLoad()
//***************************************************** end QuickLoad section

//*************************************************************** 65A section
//====================================================== 65A monitor commands
// R - reset
// L - load mode followed by <2 Byte Hex-Address><Hex-Data>, R to exit
// P - print mode followed by <2 Byte Hex-Address>
//     dumps data at address, any key to exit
// G - go with $012D=K, $012A=X, $0129=Y, $012B=A, addr=($012F lo, $012E hi)
//     via RTI. Note that unlike RTS, RTI does NOT add one to the destination
//     before placing it in the Program Counter (PC)
// 0-9
// A-F hex digit
//
// Lxxxx
//     00 11 22 33 44 55 66 77 88 99 AA BB CC DD EE FF
//
// R
//     L0120
//===========================================================================

//============================================================== WriteALoad()
// Purpose...: Write load address for 65A format
// Input.....: i_lLoadAddr: Address to write
//             i_pfOut....: File pointer to outfile
// Result....: Last write function result
//===========================================================================
int WriteALoad(long i_lLoadAddr, FILE* i_pfOut) {

  int iFnResult = 0;

  if (i_lLoadAddr > 0) {
    iFnResult = fputc( 'L', i_pfOut);

    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)((i_lLoadAddr >> 8) & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)(i_lLoadAddr & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = WriteEol(i_pfOut);
    }
  }

  return iFnResult;
}
//========================================================== eoF WriteALoad()

//================================================================ WriteAGo()
// Purpose...: Write GO for 65A format
// Input.....: i_lLoadAddr: Address to write
//             i_pfOut....: File pointer to outfile
// Remarks...: Upon G command, Serial monitor loads from the following
//             $0129 - Y, $012A - X, $012B - A
//             $012C - Proc Status 'P', $012D - stack pointer 'K'
//             $012E - PC Hi, $012F - PC Lo
//             When in Command Mode. 65A resets ACIA on every character
//             except L, P, & G
//===========================================================================
int WriteAGo(long i_lAddr, FILE* i_pfOut) {

  int iCnt;
  int iFnResult = 0;

	if (i_lAddr > 0) {

    //---------------------------------------- store address, clear registers
  	iFnResult = WriteALoad(0x0129, i_pfOut);

    //------------------------------------------------------ write Y, X, A, P
    for (iCnt = 0; ((iCnt < 4) && (iFnResult != EOF)); iCnt ++) {
      iFnResult = WriteHex(0, i_pfOut);

      if (iFnResult != EOF) {
        iFnResult = fputc(0x20, i_pfOut);
      }
    }

    //----- write stack $FD is good, since stack grows down & avoids NMI, IRQ
    iFnResult = WriteHex(0xFD, i_pfOut);

    if (iFnResult != EOF) {
      iFnResult = fputc(0x20, i_pfOut);
    }

    if (iFnResult != EOF) {
      WriteHex((BYTE)((i_lAddr >> 8) & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = fputc(0x20, i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)(i_lAddr & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = fputc('R', i_pfOut);
    }

    //-----------------------------------------------------------------------
    // the 'R' command caused the ACIA to be reset, along with
    // reinitialization of the 65A monitor.  Some characters are lost from
    // the data stream and serial syncronization may be compromised.
    // Anything other than P,G,L will also reset ACIA.
    // Since we cant pause output for a character count or two, just repeat
    // command a few times, maybe it will catch.
    //-----------------------------------------------------------------------
    for (iCnt = 0; ((iCnt < 4) && (iFnResult != EOF)); iCnt ++) {
  		iFnResult = fputc('G', i_pfOut);
    }
  }

	return (iFnResult);
}
//============================================================ eoF WriteAGo()

//================================================================ Write65A()
// Purpose...: Write file in OSI 65A serial monitor load format
// Input.....: i_pfIn.....: File pointer to infile
//             i_pfOut....: File pointer to outfile
// Result....: Last write function result
//===========================================================================
int Write65A(FILE* i_pfIn, FILE* i_pfOut) {

  int  iCnt;
  int  iNumRead = -1;
  int  iRow = 0;
  int  iFnResult = 0;
  char acBuf[m_BUFSIZE];

  //------------------------------------------------------ write load address
  iFnResult = WriteALoad(m_lLoadAddr, i_pfOut);

  while ((iFnResult != EOF) && (iNumRead != 0)) {
    iNumRead = fread(acBuf, sizeof(char), sizeof(acBuf), i_pfIn);
    iFnResult = EOF;
    
    if (iNumRead != 0) {
      iFnResult = 0;
      
      for (iCnt = 0; ((iCnt < iNumRead) && (iFnResult != EOF)); iCnt ++) {
        iFnResult = WriteHex(acBuf[iCnt], i_pfOut);
        iRow ++;

        if (iRow < 8) {

          //-------- non HEX characters are ignored, make file easier to read
          if (m_iLineEnd >= 0) {
            iFnResult = fputc(' ', i_pfOut);
          }
        }
        else {
          iRow = 0;
          iFnResult = WriteEol(i_pfOut);
        }
      }
    }
  }

  iFnResult = fputc('R', i_pfOut);

  if ((m_lGoAddr >= 0) && (iFnResult != EOF)) {

    //----------------------------------------------- no need to write loader
    if (m_lLoadAddr <= 0x12E) {

      // since we cant pause output for a character count or two, just
      // repeat command a few times, maybe it will catch

      //----------------------------- write nulls for padding after R command
      for (iCnt = 0; ((iCnt < 4) && (iFnResult != EOF)); iCnt ++) {
        iFnResult = fputc('G', i_pfOut);
      }
    }

    //--------------------- try to resync ACIA bitstream with some L commands
    else {

      for (iCnt = 0; ((iCnt < 4) && (iFnResult != EOF)); iCnt ++) {
        iFnResult = fputc('L', i_pfOut);
      }

      if (iFnResult != EOF) {
        iFnResult = WriteAGo(m_lGoAddr, i_pfOut);
      }
    }
  }

  return (iFnResult);
}
//============================================================ eoF Write65A()
//*********************************************************** end 65A section

//*************************************************************** 65V section
//====================================================== 65V monitor commands
// .    switch to address mode
// /    switch to data mode
// <CR> increment address (in data mode)
// L    set input to serial port. To transfer back, set $00FD to $00 via
//      serial or press reset or SPACE on polled keyboard
// 0-9
// A-F  hex digit
//===========================================================================

//============================================================== WriteVLoad()
// Purpose...: Write load address for 65V format
// Input.....: i_lLoadAddr: Address to write
//             i_pfOut....: File pointer to outfile
//===========================================================================
int WriteVLoad(long i_lLoadAddr, FILE* i_pfOut) {

  int iFnResult = 0;

  if (i_lLoadAddr > 0) {

    //------------------------------------------------------ set address mode
    iFnResult = fputc('.', i_pfOut); //set address mode
		
    //--------------------------------------------------------- write address
    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)((i_lLoadAddr >> 8) & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)(i_lLoadAddr & 0xFF), i_pfOut);
    }
  }

  return (iFnResult);
}
//========================================================== eoF WriteVLoad()

//================================================================ WriteVGo()
// Purpose...: Write GO for 65V format
// Input.....: i_lAddr: Address to write
//             i_pfOut: File pointer to outfile
//===========================================================================
int WriteVGo(long i_lAddr, FILE* i_pfOut) {

  int iFnResult = 0;

	if (i_lAddr > 0) {

    //--------------------------------------------------------- store address
    iFnResult = WriteVLoad(i_lAddr, i_pfOut);

    //------------------------------------------------------- execute address
    if (iFnResult != EOF) {
      iFnResult = fputc('G', i_pfOut);
    }
  }

	return (iFnResult);
}
//============================================================ eoF WriteVGo()

//================================================================ Write65V()
// Purpose...: Write 65V format
// Input.....: i_pfIn.: File pointer to infile
//             i_pfOut: File pointer to outfile
// Result....: Result of last read/write operation
//===========================================================================
int Write65V(FILE* i_pfIn, FILE* i_pfOut) {

  int  iCnt;
  int  iNumRead = -1;
  int  iFnResult = 0;
  char acBuf[m_BUFSIZE];

  //----------------------------------------------------------- store address
  iFnResult = WriteVLoad(m_lLoadAddr, i_pfOut);

  //----------------------------------------------------------- set data mode
  if (iFnResult != EOF) {
    iFnResult = fputc('/', i_pfOut);
  }
  
  while ((iFnResult != EOF) && (iNumRead != 0)) {
    iNumRead = fread(acBuf, sizeof(char), sizeof(acBuf), i_pfIn);

    for (iCnt = 0; ((iCnt < iNumRead) && (iFnResult != EOF)); iCnt ++) {
      iFnResult = WriteHex(acBuf[iCnt], i_pfOut);

      if (iFnResult != EOF) {
        iFnResult = fputc('\r', i_pfOut);
      }
    }
  } 
 
  if ((m_lGoAddr >= 0) && (iFnResult != EOF)) {
    iFnResult = WriteVGo(m_lGoAddr, i_pfOut);
  }

  return (iFnResult);
}
//============================================================ eoF Write65V()
//*********************************************************** end 65V section

//********************************************************** CsLoader section
//===================================================== MOS Technology format
// The Lines
//   Each line consists of 5 fields. These are the length field, address
//   field, data field and the checksum. The lines always start with a
//   semicolon (;) character.
//
// The Fields
//    ___ ________ _________ ______ __________ ______
//   |_;_|_Length_|_Address_|_Data_|_Checksum_|_CRLF_|
//
// Length
//   The record length field is a 2 character (1 byte) field that specifies
//   the number of data bytes in the record. Typically this is 24 or less.
//
// Address
//   This is a 2-byte address that specifies where the data in the record is
//   to be loaded into memory, big-endian.
//
// Data
//   The data field contains the executable code, memory-loadable data or
//   descriptive information to be transferred.
//
// Checksum
//   The checksum is an 2-byte field that represents the least significant
//   two bytes of the the sum of the values represented by the pairs of
//   characters making up the record length, address, and data fields,
//   big-endian.
//
// End of File (optional?)
//   The final line should have a data length of zero, and the count of data
//   lines in the address field. The checksum is not the usual checksum, it
//   is instead a repeat of the data line count.
//
// OSI specific
//  The final line consists of a $ character followed by 4 digit hex address
//  to start execution
//===========================================================================

//================================================================ WriteCGo()
// Purpose...: Write GO for MOS/OSI checksum loader
// Input.....: i_lAddr: Address to write
//             i_pfOut: File pointer to outfile
//===========================================================================
int WriteCGo(long i_lAddr, FILE* i_pfOut) {

  int iFnResult = 0;

  if (i_lAddr < 0) {

    //------------------------------------------------------ set address mode
    iFnResult = fputc('$', i_pfOut);

  	//--------------------------------------------------------- write address
    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)((i_lAddr >> 8) & 0xFF), i_pfOut);
    }

    if (iFnResult != EOF) {
      iFnResult = WriteHex((BYTE)(i_lAddr & 0xFF), i_pfOut);
    }
  }

	return (iFnResult);
}
//============================================================ eoF WriteCGo()

//================================================================= WriteCs()
// Purpose...: Write file in MOS/OSI checksum format
// Input.....: i_pfIn.: File pointer to infile
//             i_pfOut: File pointer to outfile
// Remarks...: Allowed characters for .chk are 0-9, A-G, <nul> <lf> <cr>
//             Record format:
//             ;<numbytes><adrHi><adrLo><data bytes><csHi><csLo><CR>
//===========================================================================
int WriteCs(FILE* i_pfIn, FILE* i_pfOut) {

  #define MAXROW 0x18
  
  int  iCnt;
  WORD wSum;
	unsigned long ulAddr = m_lLoadAddr;
  int  iNumRead = -1;
  int  iFnResult = 0;
  char acBuf[m_BUFSIZE];
       	
  while ((iFnResult != EOF) && (iNumRead != 0)) {
    iNumRead = fread(acBuf, sizeof(char), MAXROW, i_pfIn);
    	
		//----------------------------------------------- ;<addr><data><cksm><CR>
   	if (iNumRead > 0) {
		  iFnResult = fputc(';', i_pfOut);
      wSum = iNumRead;

		  if (iFnResult != EOF) {
        iFnResult = WriteHex((BYTE)iNumRead, i_pfOut);
      }

      //------------------------------------------------------- write address
      if (iFnResult != EOF) {
        wSum += (BYTE)((ulAddr >> 8) & 0xFF);
        iFnResult = WriteHex((BYTE)((ulAddr >> 8) & 0xFF), i_pfOut);
      }
      
      if (iFnResult != EOF) {
        wSum += (BYTE)(ulAddr & 0xFF);
        iFnResult = WriteHex((BYTE)(ulAddr & 0xFF), i_pfOut);
		  }

      ulAddr += iNumRead;
    	
      for (iCnt = 0; ((iCnt < iNumRead) && (iFnResult != EOF)); iCnt ++) {
        iFnResult = WriteHex(acBuf[iCnt], i_pfOut);
        wSum += (BYTE)acBuf[iCnt];
      }

      if (iFnResult != EOF) {
        iFnResult = WriteHex((BYTE)((wSum >> 8) & 0xFF), i_pfOut);
      }
      
      if (iFnResult != EOF) {
        iFnResult = WriteHex((BYTE)(wSum  & 0xFF), i_pfOut);
      }
      
    	//------------------------------------------------------ <CR> ends data
      if (iFnResult != EOF) {  
        fputc('\r', i_pfOut);
		  }
      
		  //------------------------------------------------------------ padding?
		  for (iCnt = 0; ((iCnt < m_iNullPadding)&&(iFnResult != EOF)); iCnt ++){
        iFnResult = fputc('\x0', i_pfOut);
		  }

		  //----------------------------------------------- <LF> starts next line
      if (iFnResult != EOF) {
        fputc('\n', i_pfOut);
      }
    }
  }

  //--------------------------------------------------- set execute addr mode
  if ((m_lGoAddr >= 0) && (iFnResult != EOF)) {
    iFnResult = fputc('$', i_pfOut);
  }

  //-------------------------------------------------------- write go address
  if (iFnResult != EOF) {
    iFnResult = WriteHex((BYTE)((m_lGoAddr >> 8) & 0xFF), i_pfOut);
  }

  if (iFnResult != EOF) {
    iFnResult = WriteHex((BYTE)(m_lGoAddr & 0xFF), i_pfOut);
  }

  return (iFnResult);
}
//============================================================= eoF WriteCs()
//****************************************************** end CsLoader section

//************************************************** checksum loaders section
//=========================================================== Checksum Loader
BYTE abLoaderCodeCs[] = {
  0x00, 0x00,                          // 2x .BYTE $00,  function unknown 
  0x00, 0x00, 0x00,                    // variables                 

  //---------------------------------------------- START: program entry point
  0x20, 0xA0, 0x00,                    // JSR GETACIA    ->patch 0x06, A0,07
  0xC9, 0x3B,                          // CMP #$3B
  0xF0, 0x13,                          // BEQ RECSTART
  0xC9, 0x24,                          // CMP #$24
  0xD0, 0xF5,                          // BNE START
                    
  //---------------------------------------------------------- get go address
  0x20, 0x81, 0x00,                    // JSR GETBYTE    ->patch 0x11, 81 07
  0x8D, 0x03, 0x00,                    // STA SPBUFHI    ->patch 0x14, 03 07
  0x20, 0x81, 0x00,                    // JSR GETBYTE    ->patch 0x17, 81 07
  0x8D, 0x02, 0x00,                    // STA SPBUFLO    ->patch 0x1A, 02 07
  0x4C, 0xAE, 0x00,                    // JMP GOMSG      ->patch 0x1D, AE 07

  //--------------------------- RECSTART: record start character was received
  0xA9, 0x00,                          // LDA #$00
  0x8D, 0x02, 0x00,                    // STA SPBUFLO    ->patch 0x22, 02 07
  0x8D, 0x03, 0x00,                    // STA SPBUFHI    ->patch 0x25, 03 07
  0x20, 0x6E, 0x00,                    // JSR RDBYTE     ->patch 0x28, 6E 07
  0xAA,                                // TAX
  0x20, 0x6E, 0x00,                    // JSR RDBYTE     ->patch 0x2C, 6E 07
  0x8D, 0x3C, 0x00,                    // STA RECSTORE+2 ->patch 0x2F, 3C 07
  0x20, 0x6E, 0x00,                    // JSR RDBYTE     ->patch 0x32, 6E 07
  0x8D, 0x3B, 0x00,                    // STA RECSTORE+1 ->patch 0x35, 3B 07

  //--------------------------------------- RECBYTE: read and store data byte
  0x20, 0x6E, 0x00,                    // JSR RDBYTE     ->patch 0x38, 6E 07
  0x8D, 0xFF, 0x0F,                    // STA $0FFF (addr=self modified code)
  0xEE, 0x3B, 0x00,                    // INC RECSTORE+1 ->patch 0x3E, 3B 07
  0xD0, 0x03,                          // BNE RECNEXT
  0xEE, 0x3C, 0x00,                    // INC RECSTORE+2 ->patch 0x43, 3C 07
  0xCA,                                // DEX
  0xD0, 0xEF,                          // BNE RECBYTE
  0x20, 0x81, 0x00,                    // JSR GETBYTE    ->patch 0x49, 81 07
  0xCD, 0x03, 0x00,                    // CMP SPBUFHI    ->patch 0x4C, 03 07
  0xD0, 0x08,                          // BNE CSERR
  0x20, 0x81, 0x00,                    // JSR GETBYTE    ->patch 0x51, 81 07
  0xCD, 0x02, 0x00,                    // CMP SPBUFLO    ->patch 0x54, 02 07
  0xF0, 0xAD,                          // BEQ START
                  
  //------------------------------------- CSERR: print checksum error message
  0xA0, 0x00,                          // LDY #$00
  0xB9, 0xD3, 0x00,                    // LDA ERRMSG,Y   ->patch 0x5B, D3 07
  0xF0, 0x06,                          // BEQ WAITKEY
  0x20, 0xEE, 0xFF,                    // JSR OUTVEC
  0xC8,                                // INY
  0xD0, 0xF5,                          // BNE CELOOP
   
  //--------------------------------------- WAITKEY: wait for key after error
  0x20, 0xA3, 0x00,                    // JSR INBYTE     ->patch 0x66, A3 07
  0xC9, 0x47,                          // CMP #$47
  0xD0, 0xF9,                          // BNE WAITKEY
  0xF0, 0x97,                          // BEQ START

  //--- RDBYTE: read byte from ACIA, update checksum in $702 and $703 ($076E)
  0x20, 0x81, 0x00,                    // JSR GETBYTE    ->patch 0x6F, 81 07
  0x18,                                // CLC
  0x6D, 0x02, 0x00,                    // ADC SPBUFLO    ->patch 0x73, 02 07
  0x8D, 0x02, 0x00,                    // STA SPBUFLO    ->patch 0x76, 02 07
  0x90, 0x03,                          // BCC RBDONE
  0xEE, 0x03, 0x00,                    // INC SPBUFHI    ->patch 0x7B, 03 07
  0xAD, 0x04, 0x00,                    // LDA SPINBUF    ->patch 0x7E, 04 07
  0x60,                                // RTS
                  
  //-------------------------------- GETBYTE: read HEX pair from ACIA ($0781)
  0x20, 0x84, 0x00,                    // JSR GETHEX     ->patch 0x82, 84 07

  //------------------- GETHEX: read byte from ACIA, convert to HEX character
  0x20, 0xA0, 0x00,                    // JSR GETACIA    ->patch 0x85, A0 07
  0xC9, 0x41,                          // CMP #$41
  0x90, 0x02,                          // BCC L078D
  0xE9, 0x07,                          // SBC #$07
  0x29, 0x0F,                          // AND #$0F
  0x0A, 0x0A, 0x0A, 0x0A,              // ASL A  (4x)
  0xA0, 0x04,                          // LDY #$04
  0x2A,                                // ROL A
  0x2E, 0x04, 0x00,                    // ROL SPINBUF    ->patch 0x97, 04 07
  0x88,                                // DEY
  0xD0, 0xF9,                          // BNE L0795
  0xAD, 0x04, 0x00,                    // LDA SPINBUF    ->patch 0x9D, 04 07
  0x60,                                // RTS
                  
  //---------------------------- GETACIA: get character from keyboard or ACIA
  0xA9, 0x80,                          // LDA #$80
  0x2C,                                // .BYTE $2C
  0xA9, 0x00,                          // LDA #$00
  0x8D, 0x03, 0x02,                    // STA $0203
  0x20, 0xEB, 0xFF,                    // JSR $FFEB
  0x4C, 0xEE, 0xFF,                    // JMP OUTVEC
                   
  //--------------------------------------------- GOMSG: print ready message
  0x20, 0x59, 0xFE,                    // JSR SCNCLR
  0xB9, 0xC4, 0x00,                    // LDA RDYMSG,Y   ->patch 0xB2, C4 07
  0xF0, 0x06,                          // BEQ GO
  0x20, 0xEE, 0xFF,                    // JSR OUTVEC
  0xC8,                                // INY
  0xD0, 0xF5,                          // BNE GMLOOP
                  
  //-------------- GO: jump to address told by $xxxx token (at end of stream)
  0xA9, 0x00,                          // LDA #$00
  0x8D, 0x03, 0x02,                    // STA $0203
  0x6C, 0x02, 0x00,                    // JMP (SPBUFLO)  ->patch 0xC2, 02 07

  //--------------------------------------------------- RDYMSG: ready message
  0x0D,                                // .BYTE $0D
                                         
  0x43, 0x73, 0x4C, 0x6F, 0x61, 0x64,  // .BYTE "CsLoad OK  "<CR><LF>
  0x20, 0x4F, 0x4B, 0x20, 0x20,
  0x0D, 0x0A, 0x00,
   
  //------------------------------------------ ERRMSG: checksum error message
  0x0D, 0x0A, 0x0A,                    // .BYTE $0D, $0A, $0A
  0x4F, 0x42, 0x4A, 0x20, 0x43, 0x48,  // .BYTE "OBJ CHECKSUM ERR"
  0x45, 0x43, 0x4B, 0x53, 0x55, 0x4D,  
  0x20, 0x45, 0x52, 0x52,        
  0x0D, 0x0A,                          // .BYTE $0D, $0A
  0x52, 0x57, 0x44, 0x20, 0x50, 0x41,  // .BYTE "RWD PAST ERR, G=RETRY"
  0x53, 0x54, 0x20, 0x45, 0x52, 0x52,  
  0x2C, 0x20, 0x47, 0x3D, 0x52, 0x45,  
  0x54, 0x52, 0x59,  
  0x00,                                // .BYTE $00
};

//========================== relocation offsets (base = $0000, start = $0005)
BYTE abRelocOffsetsCs[] = {
  0x06, 0x11, 0x14, 0x17, 0x1A, 0x1D, 0x22, 0x25,
  0x28, 0x2C, 0x2F, 0x32, 0x35, 0x38, 0x3E, 0x43,
  0x49, 0x4C, 0x51, 0x54, 0x5B, 0x66, 0x6F, 0x73,
  0x76, 0x7B, 0x7E, 0x82, 0x85, 0x97, 0x9D, 0xB2,
  0xC2
};

//============================================================== Quick Loader
//--------------------------------------- uses zero-page addresses 0xF0..0xF3
BYTE abLoaderCodeQl[] = {
  //------------------------------------------------------------- entry point
  0xEA, 0xEA, 0xEA, 0xEA, 0xEA,        // offset for fixed start (load+5)
  0x20, 0xC4, 0x00,                    // JSR CLRHOME    ->patch 0x06, C4 00
  0xD8,                                // CLD

  0x20, 0x9D, 0x00,                    // JSR RDACIA     ->patch 0x0A, 9D 00
  0xC9, 0x24,                          // CMP #$24
  0xF0, 0x54,                          // BEQ GOADDR
  0xC9, 0x3B,                          // CMP #$3B
  0xD0, 0xF5,                          // BNE START
                  
  //--------------------------- RECSTART: a ';' was received, start of record
  0xA0, 0x00,                          // LDY #$00
  0x84, 0xF0,                          // STY SPF0
  0x84, 0xF1,                          // STY SPF1
  0x20, 0x8E, 0x00,                    // JSR RDACIACS   ->patch, 0x1B 8E 00
  0xAA,                                // TAX
  0x20, 0x8E, 0x00,                    // JSR RDACIACS   ->patch, 0x1F 8E 00
  0x85, 0xF3,                          // STA SPF3
  0x20, 0xA7, 0x00,                    // JSR HEXTOCRT   ->patch, 0x24 A7 00
  0x20, 0x8E, 0x00,                    // JSR RDACIACS   ->patch, 0x27 8E 00
  0x85, 0xF2,                          // STA SPF2
  0x20, 0xA7, 0x00,                    // JSR HEXTOCRT   ->patch, 0x2C A7 00

  //---------------------------- RSLOOP: read all databytes, count given by X
  0x20, 0x8E, 0x00,                    // JSR RDACIACS   ->patch, 0x2F 8E 00
  0x91, 0xF2,                          // STA (SPF2),Y
  0xC8,                                // INY
  0xCA,                                // DEX
  0xD0, 0xF7,                          // BNE RSLOOP
                  
  //------------------------------------------------------- validate checksum
  0x20, 0x9D, 0x00,                     // JSR RDACIA    ->patch 0x38, 9D 00
  0xC5, 0xF1,                           // CMP SPF1
  0xD0, 0x15,                           // BNE ERRMSG
  0x20, 0x9D, 0x00,                     // JSR RDACIA    ->patch 0x3F, 9D 00
  0xC5, 0xF0,                           // CMP SPF0
  0xD0, 0x0E,                           // BNE ERRMSG

  //------------------------------------------------- OKMSG: print OK message
  0xBD, 0xCE, 0x00,                    // LDA MSGOK,X    ->patch 0x46, CE 00
  0xF0, 0x06,                          // BEQ JPSTART
  0x20, 0xEE, 0xFF,                    // JSR CRTOUT
  0xE8,                                // INX
  0xD0, 0xF5,                          // BNE OKMSG
  0x4C, 0x09, 0x00,                    // JMP START      ->patch 0x51, 09 00

  //-------------------------------- ERRMSG: checksum mismatch, print message
  0xBD, 0xD4, 0x00,                    // LDA MSGERR,X   ->patch 0x54, D4 00
  0xF0, 0x06,                          // BEQ RESTART
  0x20, 0xEE, 0xFF,                    // JSR CRTOUT
  0xE8,                                // INX
  0xD0, 0xF5,                          // BNE ERRMSG
  0x20, 0x00, 0xFE,                    // JSR NEWMON
  0x4C, 0x09, 0x00,                    // JMP START      ->patch 0x62, 09 00

  //-------------------------------------- GOADDR: '$' was received from ACIA
  0x20, 0x9D, 0x00,                    // JSR RDACIA     ->patch 0x65, 9D 00
  0x85, 0xF1,                          // STA SPF1
  0x20, 0x9D, 0x00,                    // JSR RDACIA     ->patch 0x6A, 9D 00
  0x85, 0xF0,                          // STA SPF0

  0x20, 0xC4, 0x00,                    // JSR CLRHOME    ->patch 0x6F, C4 00
  0xA2, 0x00,                          // LDX #$00
  0xBD, 0xDB, 0x00,                    // LDA MSGJMP,X   ->patch 0x74, DB 00
  0xF0, 0x06,                          // BEQ GO
  0x20, 0xEE, 0xFF,                    // JSR CRTOUT
  0xE8,                                // INX
  0xD0, 0xF5,                          // BNE JMPMSG

  //----------------------------------------------- GO: jump to given address
  0xA5, 0xF1,                          // LDA SPF1
  0x20, 0xA7, 0x00,                    // JSR HEXTOCRT   ->patch 0x81, A7 00
  0xA5, 0xF0,                          // LDA SPF0
  0x20, 0xA7, 0x00,                    // JSR HEXTOCRT   ->patch 0x86, A7 00
  0x20, 0xF5, 0xFB,                    // JSR CRLF
  0x6C, 0xF0, 0x00,                    // JMP ($00F0)
                  
  //------------- RDACIAS: read byte from ACIA, update checksum in SPF0, SPF1
  0x20, 0x9D, 0x00,                    // JSR RDACIA     ->patch 0x8F 9D 00
  0x48,                                // PHA
  0x18,                                // CLC
  0x65, 0xF0,                          // ADC SPF0
  0x85, 0xF0,                          // STA SPF0
  0x90, 0x02,                          // BCC RACRET
  0xE6, 0xF1,                          // INC SPF1
  0x68,                                // PLA
  0x60,                                // RTS
                  
  //---------------------------------------- RDACIA: get character from ACIA
  0xAD, 0x00, 0xF0,                    // LDA ACIACTRL
  0x4A,                                // LSR A
  0x90, 0xFA,                          // BCC RDACIA
  0xAD, 0x01, 0xF0,                    // LDA ACIADATA
  0x60,                                // RTS
                  
  //------------------ HEXTOCRT: output byte in A as two-digit HEX to screen
  0x48, 0x48,                          // PHA, 2x
  0x4A, 0x4A, 0x4A, 0x4A,              // LSR A, 4x
  0x20, 0xB8, 0x00,                    // JSR NIBTOCRT   ->patch 0xAE B8 00
  0x68,                                // PLA
  0x29, 0x0F,                          // AND #$0F
  0x20, 0xB8, 0x00,                    // JSR NIBTOCRT   ->patch 0xB4 B8 00
  0x68,                                // PLA
  0x60,                                // RTS
                 
  //------------- NIBTOCRT: output low nibble in A as HEX character to screen
  0x09, 0x30,                          // ORA #$30
  0xC9, 0x3A,                          // CMP #$3A
  0x90, 0x02,                          // BCC HCOUT
  0x69, 0x06,                          // ADC #$06
  0x20, 0xEE, 0xFF,                    // JSR CRTOUT
  0x60,                                // RTS

  //----------------------------- clear screen, home cursor (CEGMON required)
  0x20, 0xD1, 0xFF,                    // JSR CURHOM
  0x8E, 0x00, 0x02,                    // STX $0200
  0x20, 0x59, 0xFE,                    // JSR CLRSCR
  0x60,                                // RTS
  
  //------------------------------------------------------- MSGOK: OK message
  0x20, 0x4F, 0x4B,                    // .BYTE " OK"
  0x0A, 0x0D, 0x00,                    // .BYTE $0A,$0D,$00
                  
  //--------------------------------------------------- MSGERR: error message
  0x20, 0x45, 0x52, 0x52,              // .BYTE " ERR"
  0x0A, 0x0D, 0x00,                    // .BYTE $0A,$0D,$00
                  
  //------------------------------------------------------------ jump message
  0x0A, 0x0D,                          // .BYTE $0A,$0D
  0x4A, 0x4D, 0x50, 0x20, 0x24,        // .BYTE "JMP $"
  0x00,                                // .BYTE $00
                  
  0x24, 0x24, 0x24, 0x24               // .BYTE "$$$$"
};

BYTE abRelocOffsetsQl[] = {
  0x06, 0x0A, 0x1B, 0x1F, 0x24, 0x27,
  0x2C, 0x2F, 0x38, 0x3F, 0x46, 0x51, 
  0x54, 0x62, 0x65, 0x6A, 0x6F, 0x74, 
  0x81, 0x86, 0x8F, 0xAE, 0xB4
};

//=========================================================== WriteCsLoader()
// Purpose...: Write checksum loader
// Input.....: i_iTargetAddr: Target address
//             i_pfOut......: File pointer to output file
//             i_bType......: 1= checksum loader, 2= quick loader
//===========================================================================
int WriteCsLoader(long i_lTargetAddr, FILE* i_pfOut, BYTE i_bType) {

  int iCnt;
  int iCodeSize = 0;
  int iRelocSize = 0;
  short sAddr;
  BYTE* pbCode;
  BYTE* pbReloc;
	BYTE* pbBuf = NULL;
  int iFnResult = EOF;

  if (i_bType == m_LOADER_TYPE_CS) {
    iCodeSize = sizeof(abLoaderCodeCs);  
    pbCode = (BYTE*)&abLoaderCodeCs;
    iRelocSize = sizeof(abRelocOffsetsCs);
    pbReloc = (BYTE*)&abRelocOffsetsCs;
  }
  else if (i_bType == m_LOADER_TYPE_QL) {
    iCodeSize = sizeof(abLoaderCodeQl);  
    pbCode = (BYTE*)&abLoaderCodeQl;
    iRelocSize = sizeof(abRelocOffsetsQl);
    pbReloc = (BYTE*)&abRelocOffsetsQl;
  }

  if (iCodeSize != 0) {
    pbBuf = (BYTE*)malloc(iCodeSize + 1);
  }    

	if (pbBuf) {

		//----------------------------------- patch chk loader with reloc offsets
		memcpy(pbBuf, pbCode, iCodeSize);

		for (iCnt = 0; iCnt < iRelocSize; iCnt ++) {
      pbCode = pbBuf + pbReloc[iCnt];
      sAddr = *pbCode + (*(pbCode+1) << 8);
      sAddr += i_lTargetAddr;
      *pbCode = sAddr & 0xFF;
      *(pbCode+1) = sAddr >> 8;
		}

		//--------------------------------------------------------- store address
		iFnResult = WriteVLoad(i_lTargetAddr, i_pfOut);

    //--------------------------------------------------------- set data mode
		iFnResult = fputc('/', i_pfOut);

    //------------------------------------------------- write chk loader body
    for (iCnt=0; ((iCnt < iCodeSize) && (iFnResult!=EOF)); iCnt++) {
      iFnResult = WriteHex(pbBuf[iCnt], i_pfOut);

      if (iFnResult != EOF) {
        iFnResult = fputc('\r', i_pfOut);
      }
   	}

    //------------------------------------------------------ write go address
		if (iFnResult != EOF) {
      iFnResult = WriteVGo(i_lTargetAddr+5, i_pfOut);
    }

		free(pbBuf);
	}

	return (iFnResult);
}
//======================================================= eoF WriteCsLoader()
//********************************************** end checksum loaders section

//************************************************************** main program
//==================================================================== main()
int main (int i_iArgc, char* i_apcArgv[]) {

  #define MODE_A  1
  #define MODE_V  2
  #define MODE_C  3
  #define MODE_I  4
  #define MODE_Q  5
  #define MODE_O  6

  int   iArgs;
  int   iHuh = 0;
  int   iMode = 0;
  int   iChkLdrCs = -1;
  int   iChkLdrQl = -1;
  int   iOverWrite = 0;
  int   iContinue = 1;
  char  cArg;
  char* pcInFilename = NULL;
  char* pcOutFilename = NULL;
  FILE* pfIn = NULL;
  FILE* pfOut = NULL;

  if (i_iArgc == 1) {
    Help();
    return 0;
  }

  //------------------------------------------------------ parse command line
  for (iArgs = 1; iArgs < i_iArgc; iArgs ++) {

    if (*i_apcArgv[iArgs] == '-') {
      cArg = *(i_apcArgv[iArgs]+1);

      //------------------------------------------------- parse switches (-x)
      switch (tolower(cArg)) {

        //------------------------------------------------------ load address
        case 'l':
          m_lLoadAddr = CnvHexStrToBin(i_apcArgv[iArgs]+2, 4);
        break;

        //--------------------------------------------------- execute address
        case 'g': case 'h':
          m_lGoAddr = CnvHexStrToBin(i_apcArgv[iArgs]+2, 4);
        break;

        //------------------------- binary to OSI 65V ROM Monitor Load Format
        case 'v':
          iMode = MODE_V;
        break;

        //---------------------- binary to OSI 65A Serial Monitor Load Format
        case 'a':
          iMode = MODE_A;
        break;

        //---------------------------- binary to MOS/OSI Checksum Load Format
        case 'c':
          iMode = MODE_C;
        break;

        //---------------------------- binary to OSI QSave Binary Load Format
        case 'q':
          if (*(i_apcArgv[iArgs]+2) != 0) {
            iChkLdrQl = CnvHexStrToBin(i_apcArgv[iArgs]+2, 4);
          }
          iMode = MODE_Q;
        break;

        //----------------------- 65V ROM Monitor Load Format to QSave Format
        case 'o':
          if (*(i_apcArgv[iArgs]+2) != 0) {
            iChkLdrQl = CnvHexStrToBin(i_apcArgv[iArgs]+2, 4);
          }
          iMode = MODE_O;
        break;

        //----------------------- use CR/LF at end of line (normally just CR)
        case 't':
          m_iLineEnd = 1;

          if (*(i_apcArgv[iArgs]+2) == '0') {
            m_iLineEnd= -1;
          }
        break;

        //----------- number of bytes of null padding at end of line (in hex)
        case 'p':
          m_iNullPadding = CnvHexStrToBin(i_apcArgv[iArgs]+2, 2);
        break;

        //--------------------------------- hybrid 65V / Checksum load format
        case 'i':
          if (*(i_apcArgv[iArgs]+2) != 0) {
            iChkLdrCs = CnvHexStrToBin(i_apcArgv[iArgs]+2, 4);
          }
          iMode = MODE_I;
        break;

        //------------------------------------------ file overwite permission
        case 'f':
          iOverWrite = 1;
        break;

        //---------------------------------------------------- invalid option
        default:
          iHuh = 1;
        break;
      }
    }

    //------------------------------------------------ parse in/out filenames
    else {

      if (pcInFilename) {
        if (! pcOutFilename) {
          pcOutFilename = i_apcArgv[iArgs];
        }
      }
      else {
        pcInFilename = i_apcArgv[iArgs];
      }
    }
  }

  //----------------------------------------------- invalid option, show help
  if (iHuh) {
    Help();
    iContinue = 0;
  }

  //----------------------------------------------- check if action specified
  else if (! iMode) {
    fputs("\nNo action specified, use one of -[avciqo]\n", stderr);
    iContinue = -1;
  }
  
  //--------------------------------------------------------- open input file
  if (iContinue == 1) {
 
    if (! pcInFilename) {
      pfIn = stdin;

      //------------------------ ensure our binary \n's don't convert to \r\n
      #ifdef _WIN32
      _setmode(fileno(ifp), _O_BINARY);
      #endif
    }
    else {
      pfIn = fopen(pcInFilename, "rb");
    }

    if (! pfIn) {
      fputs("\nError opening input file for read\n", stderr);
      iContinue = -1;
    }
  }

  //-------------------------------------------------------- open output file
  if (iContinue == 1) {

    if (! pcOutFilename) {
      pfOut = stdout;

      //------------------------ ensure our binary \n's don't convert to \r\n
      #ifdef _WIN32
        _setmode(fileno(stdout), _O_BINARY);
      #endif
    }

    //------------------------------------------------- check if files exists
    else {
      pfOut = fopen(pcOutFilename, "rb");

      if (pfOut) {
        fclose(pfOut);

        if (! iOverWrite) {
          fprintf(stderr, "Output file %s exists, "
                          "use -f to force overwrite\r\n", pcOutFilename);
          iContinue = -1;
        }
      }
    }
    
    if ((iContinue == 1) && (pfOut != stdout)) {
      pfOut = fopen(pcOutFilename, "wb");

      if (! pfOut) {
        fputs("\nError opening output file\n", stderr);

        if ((pfIn != NULL) && (pfIn != stdin)) {
          fclose(pfIn);
          pfIn = NULL;
        }

        iContinue = -1;
      }
    }
  }
  
  //----------------------------------------------- execute selected function
  if (iContinue == 1) {
  
    if ((iMode == MODE_C) && (iChkLdrCs >= 0)) {
      iMode = MODE_I;
    }

    if (m_lGoAddr < 0) {
      m_lGoAddr = m_lLoadAddr;
    }

    switch (iMode) {

      case MODE_A:
        Write65A(pfIn, pfOut);
      break;

      case MODE_V:
        Write65V(pfIn, pfOut);
      break;

      case MODE_C:
        WriteCs(pfIn, pfOut);
      break;

      case MODE_Q:
        if (iChkLdrQl <= 0) {
          iChkLdrQl = m_lLoadAddr - 256;

          if (m_lLoadAddr < 0x0200) {
            m_lLoadAddr = 0x1F00;
          }
        }
        WriteCsLoader(iChkLdrQl, pfOut, m_LOADER_TYPE_QL);
        WriteQSave(pfIn, pfOut);
      break;

      case MODE_O:
        if (iChkLdrQl <= 0) {
          iChkLdrQl = m_lLoadAddr - 256;

          if (m_lLoadAddr < 0x0200) {
            m_lLoadAddr = 0x1F00;
          }
        }

        WriteCsLoader(iChkLdrQl, pfOut, m_LOADER_TYPE_QL);
        LodToQuickLoad(pfIn, pfOut);
      break;

      case MODE_I:
        if (iChkLdrCs <= 0) {
          iChkLdrCs = m_lLoadAddr - 256;

          if (m_lLoadAddr < 0x0200) {
            m_lLoadAddr = 0x1F00;
          }
        }

        WriteCsLoader(iChkLdrCs, pfOut, m_LOADER_TYPE_CS);
        WriteCs(pfIn, pfOut);
      break;
    }
  }

  //--------------------------------------------- close input and outut files
  if ((pfOut != NULL) && (pfOut != stdout)) {
    fclose(pfOut);
  }

  if ((pfIn != stdin) && (pfIn != NULL)) {
    fclose(pfIn);
  }

  return (iContinue);
}
//================================================================ eoF main()

//############################################################# EOF OsiALod.c